﻿using Android.App;
using Android.Widget;
using Android.OS;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Android;
using PalmSens.Techniques;
using PalmSens.Devices;
using PalmSens.Core.Simplified.Data;
using PalmSens;
using System.Threading.Tasks;
using System;
using Android.Views;
using Android.Support.V4.App;
using Android;
using Android.Content.PM;
using Android.Runtime;
using Android.Support.Design.Widget;

namespace PSSDKBiPotAuxExample
{
    [Activity(Label = "PSSDKBiPotAuxExample", MainLauncher = true, Icon = "@drawable/icon")]
    public class MainActivity : Activity
    {
        /// <summary>
        /// The psCommSimpleAndroid control that allows you to control your PalmSens/EmStat device
        /// </summary>
        private PSCommSimpleAndroid _psCommSimpleAndroid;

        /// <summary>
        /// The main view
        /// </summary>
        private View _view;

        /// <summary>
        /// List of permission ids used in the OnRequestPermissionResult
        /// </summary>
        private enum PermissionGroupIDs
        {
            BlueTooth = 0,
            ExternalStorage = 1
        }

        /// <summary>
        /// The refresh devices button
        /// </summary>
        private Button _btnRefresh;

        /// <summary>
        /// The (dis)connect button
        /// </summary>
        private Button _btnConnect;

        /// <summary>
        /// The start/abort measurement button
        /// </summary>
        private Button _btnMeasureBiPot;

        /// <summary>
        /// The start sequential multiplexer measurement button
        /// </summary>
        private Button _btnMeasureAux;

        /// <summary>
        /// The potential textview
        /// </summary>
        private TextView _txtPotential;

        /// <summary>
        /// The current textview
        /// </summary>
        private TextView _txtCurrent;

        /// <summary>
        /// The extra value textview
        /// </summary>
        private TextView _txtExtraValue;

        /// <summary>
        /// The status textview
        /// </summary>
        private TextView _txtStatus;

        /// <summary>
        /// The measurement textview
        /// </summary>
        private TextView _txtResults;

        /// <summary>
        /// The BiPot potential edittext
        /// </summary>
        private EditText _etBiPotPotential;

        /// <summary>
        /// The connected devices spinner 
        /// </summary>
        private Spinner _spinnerConnectedDevices;

        /// <summary>
        /// The extra value type spinner 
        /// </summary>
        private Spinner _spinnerExtraValue;

        /// <summary>
        /// The BiPot mode spinner 
        /// </summary>
        private Spinner _spinnerBiPotMode;

        /// <summary>
        /// The connected devices list adapter 
        /// </summary>
        private ArrayAdapter<string> _adapterConnectedDevices;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters with BiPot enabled
        /// </summary>
        private AmperometricDetection _methodCABiPot;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters with Aux Recording enabled
        /// </summary>
        private AmperometricDetection _methodCAAux;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement = null;

        /// <summary>
        /// The extra value curve
        /// </summary>
        private SimpleCurve _extraValueCurve = null;

        /// <summary>
        /// Called when [create].
        /// </summary>
        /// <param name="bundle">The bundle.</param>
        protected override void OnCreate(Bundle bundle)
        {
            base.OnCreate(bundle);

            // Set our view from the "main" layout resource
            SetContentView (Resource.Layout.Main);

            _view = FindViewById(Resource.Id.mainLayout);

            //Get reference to psCommSimpleAndroid control
            _psCommSimpleAndroid = FindViewById<PSCommSimpleAndroid>(Resource.Id.pSCommSimpleAndroid);
            _psCommSimpleAndroid.ReceiveStatus += _psCommSimpleAndroid_ReceiveStatus;
            _psCommSimpleAndroid.StateChanged += _psCommSimpleAndroid_StateChanged;
            _psCommSimpleAndroid.MeasurementStarted += _psCommSimpleAndroid_MeasurementStarted;
            _psCommSimpleAndroid.MeasurementEnded += _psCommSimpleAndroid_MeasurementEnded;
            _psCommSimpleAndroid.SimpleCurveStartReceivingData += _psCommSimpleAndroid_SimpleCurveStartReceivingData;
            _psCommSimpleAndroid.Disconnected += _psCommSimpleAndroid_Disconnected;

            //Get references to spinner controls
            _spinnerConnectedDevices = FindViewById<Spinner>(Resource.Id.spinnerConnectedDevices);
            _adapterConnectedDevices = new ArrayAdapter<string>(this, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerConnectedDevices.Adapter = _adapterConnectedDevices;
            _spinnerExtraValue = FindViewById<Spinner>(Resource.Id.spinnerExtraValue);
            ArrayAdapter<Java.Lang.ICharSequence> extraValueAdapter = ArrayAdapter<Java.Lang.ICharSequence>.CreateFromResource(this, Resource.Array.extraValue, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerExtraValue.Adapter = extraValueAdapter;
            _spinnerBiPotMode = FindViewById<Spinner>(Resource.Id.spinnerBiPotMode);
            ArrayAdapter<Java.Lang.ICharSequence> biPotModeAdapter = ArrayAdapter<Java.Lang.ICharSequence>.CreateFromResource(this, Resource.Array.bipotModes, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerBiPotMode.Adapter = biPotModeAdapter;

            //Get references to button controls
            _btnRefresh = FindViewById<Button>(Resource.Id.btnRefresh);
            _btnRefresh.Click += _btnRefresh_Click;
            _btnConnect = FindViewById<Button>(Resource.Id.btnConnect);
            _btnConnect.Click += _btnConnect_Click;
            _btnMeasureBiPot = FindViewById<Button>(Resource.Id.btnMeasureBiPot);
            _btnMeasureBiPot.Click += _btnMeasureBiPot_Click;
            _btnMeasureAux = FindViewById<Button>(Resource.Id.btnMeasureAux);
            _btnMeasureAux.Click += _btnMeasureAux_Click;

            //Get references to textview controls
            _txtPotential = FindViewById<TextView>(Resource.Id.txtPotential);
            _txtCurrent = FindViewById<TextView>(Resource.Id.txtCurrent);
            _txtExtraValue = FindViewById<TextView>(Resource.Id.txtExtraValue);
            _txtStatus = FindViewById<TextView>(Resource.Id.txtStatus);
            _txtResults = FindViewById<TextView>(Resource.Id.txtResults);

            //Get reference to edittext control
            _etBiPotPotential = FindViewById<EditText>(Resource.Id.etBiPotPotential);
            _etBiPotPotential.Text = "1";

            InitCAMethod(); //Create the linear sweep voltammetry method that defines the measurement parameters
            InitResultsListView();
        }

        /// <summary>
        /// Called after <c><see cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" /></c> &amp;mdash; or after <c><see cref="M:Android.App.Activity.OnRestart" /></c> when
        /// the activity had been stopped, but is now again being displayed to the
        /// user.
        /// </summary>
        /// <remarks>
        /// <para tool="javadoc-to-mdoc">Called after <c><see cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" /></c> &amp;mdash; or after <c><see cref="M:Android.App.Activity.OnRestart" /></c> when
        /// the activity had been stopped, but is now again being displayed to the
        /// user.  It will be followed by <c><see cref="M:Android.App.Activity.OnResume" /></c>.
        /// </para>
        /// <para tool="javadoc-to-mdoc">
        ///   <i>Derived classes must call through to the super class's
        /// implementation of this method.  If they do not, an exception will be
        /// thrown.</i>
        /// </para>
        /// <para tool="javadoc-to-mdoc">
        ///   <format type="text/html">
        ///     <a href="http://developer.android.com/reference/android/app/Activity.html#onStart()" target="_blank">[Android Documentation]</a>
        ///   </format>
        /// </para>
        /// </remarks>
        /// <since version="Added in API level 1" />
        /// <altmember cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" />
        /// <altmember cref="M:Android.App.Activity.OnStop" />
        /// <altmember cref="M:Android.App.Activity.OnResume" />
        protected override void OnStart()
        {
            base.OnStart();
            //Set the current app context in the psCommSimpleAndroid control
            _psCommSimpleAndroid.CurrentContext = this;
        }

        /// <summary>
        /// Called when app process is resumed and after OnStart when app is started
        /// </summary>
        protected override void OnResume()
        {
            base.OnResume();

            RequestDangerousPermissions(); //Necessary as of Android version 23
        }

        /// <summary>
        /// Request permissions that require user confirmation since android 23
        /// </summary>
        private void RequestDangerousPermissions()
        {
            //Only request permissions on Android versions after 22
            if (Build.VERSION.SdkInt >= BuildVersionCodes.M)
            {
                //Check if location permission is granted, required for BlueTooth
                if (ActivityCompat.CheckSelfPermission(this, Manifest.Permission.AccessFineLocation) != (int)Permission.Granted)
                {
                    //Request location permission
                    ActivityCompat.RequestPermissions(this, new string[] { Manifest.Permission.AccessFineLocation }, (int)PermissionGroupIDs.BlueTooth);
                }
            }
        }

        /// <summary>
        /// Callback received when user permission request has been completed
        /// </summary>
        /// <param name="requestCode"></param>
        /// <param name="permissions"></param>
        /// <param name="grantResults"></param>
        public override void OnRequestPermissionsResult(int requestCode, string[] permissions, [GeneratedEnum] Permission[] grantResults)
        {
            switch ((PermissionGroupIDs)requestCode)
            {
                case PermissionGroupIDs.BlueTooth:
                    if (grantResults.Length == 1 && grantResults[0] == Permission.Granted)
                    {
                        //Permission granted
                        if (!_psCommSimpleAndroid.Connected)
                            DiscoverConnectedDevices(); //Search for BlueTooth devices
                    }
                    else
                    {
                        //Permission denied
                        if (ActivityCompat.ShouldShowRequestPermissionRationale(this, Manifest.Permission.AccessFineLocation))
                        {
                            //Inform user that BlueTooth 
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.", Snackbar.LengthLong).Show();
                        }
                        else
                        {
                            //Occurs when user previously denied permission and selected never ask again
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.\nTo use BlueTooth you need to manualy grant the fine location for this app in the android settings", Snackbar.LengthLong).Show();
                        }
                    }
                    break;
                case PermissionGroupIDs.ExternalStorage: //Not used in this example
                default:
                    base.OnRequestPermissionsResult(requestCode, permissions, grantResults);
                    break;
            }
        }

        /// <summary>
        /// Initializes the data grid view control.
        /// </summary>
        private void InitResultsListView()
        {
            _txtResults.Text = "ID     Time (s)   Current (µA)      Extra Value";
        }

        /// <summary>
        /// Initializes the CA method.
        /// </summary>
        private void InitCAMethod()
        {
            //BiPot Method
            _methodCABiPot = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCABiPot.Potential = 1f; //Sets the potential at 1V
            _methodCABiPot.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCABiPot.RunTime = 3f; //Sets the measurement runtime at 3s

            _methodCABiPot.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCABiPot.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCABiPot.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCABiPot.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA

            //BiPot settings
            _methodCABiPot.ExtraValueMsk = ExtraValueMask.BipotWE;
            //The BiPot mode and potential are updated with the settings specified in the Form in btnMeasureBiPot_Click
            _methodCABiPot.BipotModePS = Method.EnumPalmSensBipotMode.offset;
            _methodCABiPot.BiPotPotential = 0;
            _methodCABiPot.BiPotCR = new CurrentRange(CurrentRanges.cr1uA); //Set BiPot to measure in the 1µA current range
            
            //Init EmStat Pico settings, these do not affect other devices (alternatively use the method _methodCABiPot.DeterminePGStatMode((EmStatPicoCapabilities)_psCommSimpleAndroid.Capabilities) to determine the recommended settings for the EmStat Pico)
            _methodCABiPot.PGStatMode = MethodScript.PGStatModes.LowSpeed; //BiPot measurements require the device to be set to lowspeed mode
            _methodCABiPot.SelectedPotentiostatChannel = PotentionstatChannels.Ch0; //In BiPot measurements the main we, the re and ce are controlled by channel 0 and the BiPot we by channel 1
            
            //Aux Method
            _methodCAAux = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCAAux.Potential = 1f; //Sets the potential at 1V
            _methodCAAux.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCAAux.RunTime = 3f; //Sets the measurement runtime at 3s

            _methodCAAux.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCAAux.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCAAux.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCAAux.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA

            //Aux settings
            _methodCAAux.ExtraValueMsk = ExtraValueMask.AuxInput;
        }

        /// <summary>
        /// Displays a Toast message.
        /// </summary>
        /// <param name="message">The message.</param>
        private void DisplayMessage(string message)
        {
            Toast toast = Toast.MakeText(this, message, ToastLength.Short);
            toast.Show();
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the spinner control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            _btnRefresh.Click -= _btnRefresh_Click;
            _btnRefresh.Text = "Refreshing...";
            _btnRefresh.Enabled = false;
            _adapterConnectedDevices.Clear();
            DisplayMessage($"Searching for available devices.");

            try
            {
                _connectedDevices = await _psCommSimpleAndroid.GetConnectedDevices(10000); //Discover connected devices
            }
            catch (Exception ex)
            {
                DisplayMessage(ex.Message);
            }

            foreach (Device d in _connectedDevices)
                _adapterConnectedDevices.Add(d.ToString()); //Add connected devices to spinner control

            int nDevices = _connectedDevices.Length;

            DisplayMessage($"Found {nDevices} device(s).");

            _btnConnect.Enabled = nDevices > 0;
            _btnRefresh.Text = "Refresh";
            _btnRefresh.Enabled = true;
            _btnRefresh.Click += _btnRefresh_Click;
        }

        /// <summary>
        /// Handles the Click event of the _btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private async void _btnRefresh_Click(object sender, System.EventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private async void _btnConnect_Click(object sender, System.EventArgs e)
        {
            _btnConnect.Enabled = false;
            if (!_psCommSimpleAndroid.Connected) //Determine whether a device is currently connected
            {
                if (_adapterConnectedDevices.Count == 0)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    await _psCommSimpleAndroid.Connect(_connectedDevices[_spinnerConnectedDevices.SelectedItemPosition]);
                    DisplayMessage($"Connected to {_psCommSimpleAndroid.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
                finally
                {
                    //Update UI based on connection status
                    _spinnerConnectedDevices.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnRefresh.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnConnect.Text = _psCommSimpleAndroid.Connected ? "Disconnect" : "Connect";
                    _btnMeasureAux.Enabled = _psCommSimpleAndroid.Connected;

                    //Check device capabilities to see if BiPot is supported
                    _btnMeasureBiPot.Enabled = _psCommSimpleAndroid.Connected && _psCommSimpleAndroid.Capabilities.BiPotPresent;
                    _spinnerExtraValue.Enabled = _psCommSimpleAndroid.Connected && _psCommSimpleAndroid.Capabilities.BiPotPresent;
                    _spinnerExtraValue.SetSelection((_psCommSimpleAndroid.Connected && _psCommSimpleAndroid.Capabilities.BiPotPresent) ? 0 : 1);
                }
            }
            else
            {
                await _psCommSimpleAndroid.Disconnect(); //Disconnect from the connected device
            }
            _btnConnect.Enabled = true;
        }

        private async void _btnMeasureAux_Click(object sender, System.EventArgs e)
        {
            _btnMeasureAux.Enabled = _btnMeasureBiPot.Enabled = false;
            if (_psCommSimpleAndroid.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    InitResultsListView(); //Reset the data listview
                    //Start measurement with auxillary readings enabled defined in the method
                    _activeMeasurement = await _psCommSimpleAndroid.Measure(_methodCAAux);

                    //Create a new SimpleCurve with ExtraValue(BiPot/Aux) readings (Set to silent to prevent the new data event from being raised by multiple curves)
                    _extraValueCurve = (_activeMeasurement.NewSimpleCurve(PalmSens.Data.DataArrayType.Time, PalmSens.Data.DataArrayType.AuxInput, "", true))[0];
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            else
            {
                try
                {
                    await _psCommSimpleAndroid.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            _btnMeasureAux.Enabled = _btnMeasureBiPot.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the _btnMeasureBiPot control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private async void _btnMeasureBiPot_Click(object sender, System.EventArgs e)
        {
            _btnMeasureAux.Enabled = _btnMeasureBiPot.Enabled = false;
            if (_psCommSimpleAndroid.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    //Parse the BiPot potential specified in the textbox
                    float potentialBiPot;
                    if (!float.TryParse(_etBiPotPotential.Text, out potentialBiPot))
                    {
                        DisplayMessage("Could not parse BiPot potential.");
                        return;
                    }
                    _methodCABiPot.BiPotPotential = potentialBiPot; //Set the BiPot potential specified in the form
                    //Set the BiPot mode specified in the form
                    _methodCABiPot.BipotModePS = _spinnerBiPotMode.SelectedItemPosition == 0 ? Method.EnumPalmSensBipotMode.constant : Method.EnumPalmSensBipotMode.offset;

                    InitResultsListView(); //Reset the data listview
                    //Start BiPot measurement defined in the method
                    _activeMeasurement = await _psCommSimpleAndroid.Measure(_methodCABiPot);

                    //Create a new SimpleCurve with ExtraValue(BiPot/Aux) readings (Set to silent to prevent the new data event from being raised by multiple curves)
                    _extraValueCurve = (_activeMeasurement.NewSimpleCurve(PalmSens.Data.DataArrayType.Time, PalmSens.Data.DataArrayType.BipotCurrent, "", true))[0];
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            else
            {
                try
                {
                    await _psCommSimpleAndroid.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            _btnMeasureAux.Enabled = _btnMeasureBiPot.Enabled = true;
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Comm.StatusEventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_ReceiveStatus(object sender, PalmSens.Comm.StatusEventArgs e)
        {
            Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
            double potential = status.PotentialReading.Value; //Get the potential
            double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
            PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
            CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

            _txtPotential.Text = $"Potential: {potential.ToString("F3")} V";
            _txtCurrent.Text = $"Current: {currentInRange.ToString("F3")} * {cr.ToString()}";

            if (_spinnerExtraValue.SelectedItemPosition == 0)
                _txtExtraValue.Text = $"{status.GetCorrectedBiPotCurrent().ToString("F3")} * {cr.ToString()}" ;
            else
                _txtExtraValue.Text = $"{status.GetAuxInputAsVoltage().ToString("F3")} V";
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CurrentState">State of the current.</param>
        private void _psCommSimpleAndroid_StateChanged(object sender, CommManager.DeviceState CurrentState)
        {
            _txtStatus.Text = CurrentState.ToString(); //Updates the device state indicator textbox
            _btnConnect.Enabled = CurrentState == CommManager.DeviceState.Idle;
            _btnMeasureBiPot.Text = CurrentState == CommManager.DeviceState.Idle ? "Measure" : "Abort";
            _btnMeasureAux.Text = CurrentState == CommManager.DeviceState.Idle ? "Measure" : "Abort";
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_MeasurementEnded(object sender, Exception e)
        {
            DisplayMessage("Measurement ended.");
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_MeasurementStarted(object sender, EventArgs e)
        {
            DisplayMessage("Measurement started.");
        }

        /// <summary>
        /// Raised when a Simple Curve in the active SimpleMeasurement starts receiving data
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="activeSimpleCurve">The active simple curve.</param>
        private void _psCommSimpleAndroid_SimpleCurveStartReceivingData(object sender, SimpleCurve activeSimpleCurve)
        {
            if (activeSimpleCurve.YAxisDataType != PalmSens.Data.DataArrayType.Current) //Ignore other automatically generated curves
                return;

            //Subscribe to the curve's events to receive updates when new data is available and when it iss finished receiving data
            activeSimpleCurve.NewDataAdded += ActiveSimpleCurve_NewDataAdded;
            activeSimpleCurve.CurveFinished += ActiveSimpleCurve_CurveFinished;

            DisplayMessage($"Curve is receiving new data...");
        }

        /// <summary>
        /// Raised when new data points are added to an active SimpleCurve
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Data.ArrayDataAddedEventArgs"/> instance containing the event data.</param>
        private void ActiveSimpleCurve_NewDataAdded(object sender, PalmSens.Data.ArrayDataAddedEventArgs e)
        {
            if (Looper.MyLooper() != Looper.MainLooper) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                using (var h = new Handler(Looper.MainLooper)) h.Post(() => ActiveSimpleCurve_NewDataAdded(sender, e));
                return;
            }
            SimpleCurve activeCurve = sender as SimpleCurve;
            int startIndex = e.StartIndex; //The index of the first new data point added to the curve
            int count = e.Count; //The number of new data points added to the curve

            for (int i = startIndex; i < startIndex + count; i++)
            {
                double xValue = activeCurve.XAxisValue(i); //Get the value on Curve's X-Axis (potential) at the specified index
                double yValue = activeCurve.YAxisValue(i); //Get the value on Curve's Y-Axis (current) at the specified index

                //Get the BiPot/Aux reading
                double extraValue = _extraValueCurve.YAxisValue(i); //BiPot
                string extraValueUnit = _activeMeasurement.Measurement.Method.ExtraValueMsk.HasFlag(ExtraValueMask.BipotWE) ? "µA" : "V";

                _txtResults.Text += ($"\n{i + 1}     {xValue.ToString("F2")}     {yValue.ToString("E3")}     {extraValue.ToString("E3")} {extraValueUnit}");
            }
        }

        /// <summary>
        /// Raised when a SimpleCurve stops receiving new data points
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void ActiveSimpleCurve_CurveFinished(object sender, EventArgs e)
        {
            if (Looper.MyLooper() != Looper.MainLooper) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                using (var h = new Handler(Looper.MainLooper)) h.Post(() => ActiveSimpleCurve_CurveFinished(sender, e));
                return;
            }
            SimpleCurve activeCurve = sender as SimpleCurve;

            int nDataPointsReceived = activeCurve.NDataPoints;
            DisplayMessage($"{nDataPointsReceived} data point(s) received.");

            //Unsubscribe from the curves events to avoid memory leaks
            activeCurve.NewDataAdded -= ActiveSimpleCurve_NewDataAdded;
            activeCurve.CurveFinished -= ActiveSimpleCurve_CurveFinished;

            DisplayMessage($"Curve Finished");
        }

        /// <summary>
        /// Raised when a device is disconnected.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CommErrorException">The comm error exception, this only has a value when a disconnect occurred due to a communication error.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private void _psCommSimpleAndroid_Disconnected(object sender, Exception CommErrorException)
        {
            if (CommErrorException != null)
                DisplayMessage(CommErrorException.Message);
            DisplayMessage("Disconnected");

            //Update UI based on connection status
            _spinnerConnectedDevices.Enabled = true;
            _btnRefresh.Enabled = true;
            _btnConnect.Text = "Connect";
            _btnMeasureAux.Enabled = false;
            _btnMeasureBiPot.Enabled = false;
            _spinnerExtraValue.Enabled = false;
            _btnMeasureBiPot.Text = "Measure";
            _btnMeasureAux.Text = "Measure";
            _spinnerExtraValue.SetSelection(1);
        }
    }
}

