﻿using Android.App;
using Android.OS;
using Android.Support.V7.App;
using Android.Runtime;
using Android.Widget;
using System.Threading.Tasks;
using System;
using PalmSens.Core.Simplified.Android;
using PalmSens.Devices;
using Android.Support.V4.App;
using Android;
using Android.Content.PM;
using Android.Support.Design.Widget;
using Android.Views;
using Xamarin.Essentials;
using PalmSens.Comm;
using System.Collections.Generic;
using Java.IO;
using static Android.Widget.AdapterView;
using PalmSens;
using PalmSens.Data;
using System.IO;
using System.Linq;
using System.Text;
using PalmSens.Core.Simplified.InternalStorage;
using PalmSens.DataFiles;

namespace PSSDKFileSystemBasicExample
{
    [Activity(Label = "@string/app_name", Theme = "@style/AppTheme", MainLauncher = true)]
    public class MainActivity : AppCompatActivity
    {
        /// <summary>
        /// The psCommSimpleAndroid control that allows you to control your PalmSens/EmStat device
        /// </summary>
        private PSCommSimpleAndroid _psCommSimpleAndroid;

        /// <summary>
        /// The main view
        /// </summary>
        private View _view;

        /// <summary>
        /// List of permission ids used in the OnRequestPermissionResult
        /// </summary>
        /// 
        private enum PermissionGroupIDs
        {
            BlueTooth = 0,
            ExternalStorage = 1
        }

        /// <summary>
        /// The refresh devices button
        /// </summary>
        private Button _btnRefresh;

        /// <summary>
        /// The File System button
        /// </summary>
        private Button _btnFileSystem;

        /// <summary>
        /// The (dis)connect button
        /// </summary>
        private Button _btnConnect;

        /// <summary>
        /// The File System list view
        /// </summary>
        private ListView _listViewFileSystem;

        /// <summary>
        /// The connected devices spinner 
        /// </summary>
        private Spinner _spinnerConnectedDevices;

        /// <summary>
        /// The connected devices list adapter 
        /// </summary>
        private ArrayAdapter<string> _adapterConnectedDevices;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The contents of the current directory on the internal storage
        /// </summary>
        private IDictionary<int, IInternalStorageItem> _contents = new Dictionary<int, IInternalStorageItem>();

        /// <summary>
        /// The root directory on the device.
        /// </summary>
        private IInternalStorageItem _root;

        /// <summary>
        /// Raised when a device is disconnected.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CommErrorException">The comm error exception, this only has a value when a disconnect occured due to a communication error.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private void _psCommSimpleAndroid_Disconnected(object sender, Exception CommErrorException)
        {
            DisplayMessage("Disconnected");
            if (CommErrorException != null)
                DisplayMessage(CommErrorException.Message);

            //Update UI based on connection status
            _spinnerConnectedDevices.Enabled = true;
            _btnRefresh.Enabled = true;
            _btnConnect.Text = "Connect";
            _btnFileSystem.Enabled = false;
        }

        /// <summary>
        /// Called when app process is started
        /// </summary>
        protected override void OnStart()
        {
            base.OnStart();
            //Set the current app context in the psCommSimpleAndroid control
            _psCommSimpleAndroid.CurrentContext = this;
        }

        /// <summary>
        /// Called when app process is resumed and after OnStart when app is started
        /// </summary>
        protected override void OnResume()
        {
            base.OnResume();
            RequestDangerousPermissions(); //Necessary as of Android version 23
        }

        /// <summary>
        /// Request permissions that require user confirmation since android 23
        /// </summary>
        private void RequestDangerousPermissions()
        {
            //Only request permissions on Android versions after 22
            if (Build.VERSION.SdkInt >= BuildVersionCodes.M)
            {
                //Check if location permission is granted, required for BlueTooth
                if (ActivityCompat.CheckSelfPermission(this, Manifest.Permission.AccessFineLocation) != (int)Permission.Granted)
                {
                    //Request location permission
                    ActivityCompat.RequestPermissions(this, new string[] { Manifest.Permission.AccessFineLocation }, (int)PermissionGroupIDs.BlueTooth);
                }
            }
        }

        /// <summary>
        /// Callback received when user permission request has been completed
        /// </summary>
        /// <param name="requestCode"></param>
        /// <param name="permissions"></param>
        /// <param name="grantResults"></param>
        public override void OnRequestPermissionsResult(int requestCode, string[] permissions, [GeneratedEnum] Permission[] grantResults)
        {
            switch ((PermissionGroupIDs)requestCode)
            {
                case PermissionGroupIDs.BlueTooth:
                    if (grantResults.Length == 1 && grantResults[0] == Permission.Granted)
                    {
                        //Permission granted
                        if (!_psCommSimpleAndroid.Connected)
                            DiscoverConnectedDevices(); //Search for BlueTooth devices
                    }
                    else
                    {
                        //Permission denied
                        if (ActivityCompat.ShouldShowRequestPermissionRationale(this, Manifest.Permission.AccessFineLocation))
                        {
                            //Inform user that BlueTooth 
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.", Snackbar.LengthLong).Show();
                        }
                        else
                        {
                            //Occurs when user previously denied permission and selected never ask again
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.\nTo use BlueTooth you need to manualy grant the fine location for this app in the android settings", Snackbar.LengthLong).Show();
                        }
                    }
                    break;
                case PermissionGroupIDs.ExternalStorage: //Not used in this example
                default:
                    base.OnRequestPermissionsResult(requestCode, permissions, grantResults);
                    break;
            }
        }

        /// <summary>
        /// Called when [created].
        /// </summary>
        /// <param name="bundle">The bundle.</param>
        protected override void OnCreate(Bundle savedInstanceState)
        {
            base.OnCreate(savedInstanceState);
            Xamarin.Essentials.Platform.Init(this, savedInstanceState);
            // Set our view from the "main" layout resource
            SetContentView(Resource.Layout.Main);

            _view = FindViewById(Resource.Id.mainLayout);
            //Get reference to psCommSimpleAndroid control
            _psCommSimpleAndroid = FindViewById<PSCommSimpleAndroid>(Resource.Id.pSCommSimpleAndroid);
            _psCommSimpleAndroid.Disconnected += _psCommSimpleAndroid_Disconnected;

            //Get reference to spinner control
            _spinnerConnectedDevices = FindViewById<Spinner>(Resource.Id.spinnerConnectedDevices);
            _adapterConnectedDevices = new ArrayAdapter<string>(this, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerConnectedDevices.Adapter = _adapterConnectedDevices;

            //Get references to button controls
            _btnRefresh = FindViewById<Button>(Resource.Id.btnRefresh);
            _btnRefresh.Click += _btnRefresh_Click;
            _btnConnect = FindViewById<Button>(Resource.Id.btnConnect);
            _btnConnect.Click += _btnConnect_Click;
            _btnFileSystem = FindViewById<Button>(Resource.Id.btnFileSystem);
            _btnFileSystem.Click += _btnFileSystem_Click;

            //Get references to textview controls
            _listViewFileSystem = FindViewById<ListView>(Resource.Id.listView1);
            _listViewFileSystem.ItemClick += _listViewFileSystem_ItemClick;

        }

        private void DisplayError(Exception exception)
        {
            DisplayMessage($"An error occurred: {exception.Message}");
        }

        /// <summary>
        /// Displays a Toast message.
        /// </summary>
        /// <param name="message">The message.</param>
        private void DisplayMessage(string message)
        {
            Toast toast = Toast.MakeText(this, message, ToastLength.Short);
            toast?.Show();
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the spinner control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            _btnRefresh.Click -= _btnRefresh_Click;
            _btnRefresh.Text = "Refreshing...";
            _btnRefresh.Enabled = false;
            _btnFileSystem.Enabled = false;
            _adapterConnectedDevices.Clear();
            DisplayMessage($"Searching for available devices.");

            try
            {
                _connectedDevices = await _psCommSimpleAndroid.GetConnectedDevices(10000); //Discover connected devices
            }
            catch (Exception ex)
            {
                DisplayError(ex);
            }

            foreach (Device d in _connectedDevices)
                _adapterConnectedDevices.Add(d.ToString()); //Add connected devices to spinner control

            int nDevices = _connectedDevices.Length;

            DisplayMessage($"Found {nDevices} device(s).");

            _btnConnect.Enabled = nDevices > 0;
            _btnRefresh.Text = "Refresh";
            _btnRefresh.Enabled = true;
            _btnRefresh.Click += _btnRefresh_Click;
        }

        /// <summary>
        /// Handles the Click event of the _btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private async void _btnRefresh_Click(object sender, System.EventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private async void _btnConnect_Click(object sender, System.EventArgs e)
        {
            _btnConnect.Enabled = false;
            if (!_psCommSimpleAndroid.Connected) //Determine whether a device is currently connected
            {
                if (_adapterConnectedDevices.Count == 0)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    await _psCommSimpleAndroid.Connect(_connectedDevices[_spinnerConnectedDevices.SelectedItemPosition]);
                    if (_psCommSimpleAndroid.Connected)
                    {
                        DisplayMessage($"Connected to {_psCommSimpleAndroid.ConnectedDevice.ToString()}");
                    }
                }
                catch (Exception ex)
                {
                    DisplayError(ex);
                }
                finally
                {
                    //Update UI based on connection status
                    _spinnerConnectedDevices.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnRefresh.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnConnect.Text = _psCommSimpleAndroid.Connected ? "Disconnect" : "Connect";
                    _btnFileSystem.Enabled = _psCommSimpleAndroid.Connected;

                }
            }
            else
            {
                await _psCommSimpleAndroid.Disconnect(); //Disconnect from the connected device

            }

            _btnConnect.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the _btnFileSystem control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private async void _btnFileSystem_Click(object sender, EventArgs e)
        {
            try
            {
                if (_psCommSimpleAndroid.DeviceState != CommManager.DeviceState.Idle) //Determine whether the device is currently idle
                {
                    return;
                }

                _btnFileSystem.Enabled = false;

                _root = await _psCommSimpleAndroid.GetInternalStorageBrowser().GetRootAsync();

                await OpenItemAsync(_root);

                if (_psCommSimpleAndroid.DeviceState != CommManager.DeviceState.Error)
                {
                    _btnFileSystem.Enabled = true;
                }
            }
            catch (Exception ex)
            {
                DisplayError(ex);
            }
        }

        private async Task OpenItemAsync(IInternalStorageItem item)
        {
            if (item.ItemType == DeviceFileType.Folder)
            {
                await DisplayFolderItems((IInternalStorageFolder)item);
            }
            else
            {
                await ProcessMeasurement((IInternalStorageFile)item);
            }
        }

        /// <summary>
        /// Handles the Click event of FileSystem listview item
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private async void _listViewFileSystem_ItemClick(object sender, AdapterView.ItemClickEventArgs e)
        {
            _listViewFileSystem.Enabled = false;
            var item = _contents[e.Position];

            await OpenItemAsync(item);

            _listViewFileSystem.Enabled = true;
        }

        /// <summary>
        /// Process the measurement and load it from the device.
        /// </summary>
        /// <param name="item">The storage file item</param>
        /// <returns></returns>
        private async Task ProcessMeasurement(IInternalStorageFile item)
        {
            try
            {
                var measurement = await item.GetMeasurementAsync(MeasType.New);
                DisplayMessage(measurement != null ? $"Load file '{item.Name}', with method '{measurement.Method}'" : $"Could not load file '{item.Name}'");
            }
            catch (Exception)
            {
                _psCommSimpleAndroid.Comm.State = PalmSens.Comm.CommManager.DeviceState.Error;
                //after release of the semaphore, timer will write to socket and get an exception while parsing
                _psCommSimpleAndroid.Comm.PauseParsing();
            }
        }

        /// <summary>
        /// Display the folder subfolders and any files
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        private async Task DisplayFolderItems(IInternalStorageFolder item)
        {
            DisplayMessage($"Opening path: {item.Name}");

            if (item.ItemType == DeviceFileType.Measurement) return;

            var folder = (IInternalStorageFolder) item;
            var files = await folder.GetFilesAsync();
            var subFolders = await folder.GetSubFoldersAsync();
            var totalItems = files.Count + subFolders.Count;
            var hasParent = !ReferenceEquals(item, _root); 
            
            DisplayMessage($"Loading {totalItems} storage items.");

            int count = 0;
            List<string> fileList = new List<string>();

            if (hasParent)
            {
                fileList.Add($"{count}. <Back> {folder.Parent.Name}");
                _contents[count] = folder.Parent;
                count++;
            }

            count = LoadChildrenItems(fileList, subFolders, count);
            LoadChildrenItems(fileList, files, count);
            
            if (fileList.Count > 0)
            {
                UpdateListView(fileList);
            }
        }

        /// <summary>
        /// Load the child items and add to the file list.
        /// </summary>
        /// <returns></returns>
        private int LoadChildrenItems(ICollection<string> fileList, IEnumerable<IInternalStorageItem> items, int index)
        {
            foreach (var item in items.OrderBy(i => i.Name))
            {
                var displayName = CreateListInternalStorageItemText(item, index);
                fileList.Add(displayName);
                _contents[index] = item;
                index++;
            }

            return index;
        }

        /// <summary>
        /// Create the list item display text.
        /// </summary>
        /// <returns></returns>
        private string CreateListInternalStorageItemText(IInternalStorageItem item, int index)
        {
            var type = item.ItemType == DeviceFileType.Folder ? "<Dir>" : "<File>";
            return $"{index}. {type} {item.Name}";
        }

        /// <summary>
        /// Update ListView of UI
        /// </summary>
        /// <param name="fileList">list of files</param>
        private void UpdateListView(IList<string> fileList)
        {
            using (var fileAdapter = new ArrayAdapter<string>(this, Android.Resource.Layout.SimpleListItem1, fileList))
            {
                _listViewFileSystem.Adapter = fileAdapter;
            }
        }
    }
}