﻿using Android.App;
using Android.Widget;
using Android.OS;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Android;
using PalmSens.Core.Simplified.Data;
using PalmSens.Devices;
using PalmSens.Techniques;
using PalmSens;
using System.Threading.Tasks;
using System;
using Android.Views;
using Android.Support.V4.App;
using Android;
using Android.Content.PM;
using Android.Runtime;
using Android.Support.Design.Widget;

namespace PSSDKMultiplexerExample
{
    [Activity(Label = "PSSDKMultiplexerExample", MainLauncher = true, Icon = "@drawable/icon")]
    public class MainActivity : Activity
    {
        /// <summary>
        /// The psCommSimpleAndroid control that allows you to control your PalmSens/EmStat device
        /// </summary>
        private PSCommSimpleAndroid _psCommSimpleAndroid;

        /// <summary>
        /// The main view
        /// </summary>
        private View _view;

        /// <summary>
        /// List of permission ids used in the OnRequestPermissionResult
        /// </summary>
        private enum PermissionGroupIDs
        {
            BlueTooth = 0,
            ExternalStorage = 1
        }

        /// <summary>
        /// The refresh devices button
        /// </summary>
        private Button _btnRefresh;

        /// <summary>
        /// The (dis)connect button
        /// </summary>
        private Button _btnConnect;

        /// <summary>
        /// The start/abort measurement button
        /// </summary>
        private Button _btnMeasure;

        /// <summary>
        /// The start sequential multiplexer measurement button
        /// </summary>
        private Button _btnMeasureSequentially;

        /// <summary>
        /// The start alternating multiplexer measurement button
        /// </summary>
        private Button _btnMeasureAlternatingly;

        /// <summary>
        /// The potential textview
        /// </summary>
        private TextView _txtPotential;

        /// <summary>
        /// The current textview
        /// </summary>
        private TextView _txtCurrent;

        /// <summary>
        /// The status textview
        /// </summary>
        private TextView _txtStatus;

        /// <summary>
        /// The measurement textview
        /// </summary>
        private TextView _txtResults;

        /// <summary>
        /// The connected devices spinner 
        /// </summary>
        private Spinner _spinnerConnectedDevices;

        /// <summary>
        /// The multiplexer channel spinner 
        /// </summary>
        private Spinner _spinnerChannels;

        /// <summary>
        /// The connected devices list adapter 
        /// </summary>
        private ArrayAdapter<string> _adapterConnectedDevices;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters
        /// </summary>
        private AmperometricDetection _methodCA;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters 
        /// and mulitplexer settings for sequential measurements on different channels
        /// </summary>
        private AmperometricDetection _methodCASequential;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters 
        /// and mulitplexer settings for alternating measurements on different channels
        /// </summary>
        private AmperometricDetection _methodCAAlternating;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement = null;

        /// <summary>
        /// Called when [create].
        /// </summary>
        /// <param name="bundle">The bundle.</param>
        protected override void OnCreate(Bundle bundle)
        {
            base.OnCreate(bundle);

            // Set our view from the "main" layout resource
            SetContentView (Resource.Layout.Main);

            _view = FindViewById(Resource.Id.mainLayout);

            //Get reference to psCommSimpleAndroid control
            _psCommSimpleAndroid = FindViewById<PSCommSimpleAndroid>(Resource.Id.pSCommSimpleAndroid);
            _psCommSimpleAndroid.ReceiveStatus += _psCommSimpleAndroid_ReceiveStatus;
            _psCommSimpleAndroid.StateChanged += _psCommSimpleAndroid_StateChanged;
            _psCommSimpleAndroid.MeasurementStarted += _psCommSimpleAndroid_MeasurementStarted;
            _psCommSimpleAndroid.MeasurementEnded += _psCommSimpleAndroid_MeasurementEnded;
            _psCommSimpleAndroid.SimpleCurveStartReceivingData += _psCommSimpleAndroid_SimpleCurveStartReceivingData;
            _psCommSimpleAndroid.Disconnected += _psCommSimpleAndroid_Disconnected;

            //Get references to spinner controls
            _spinnerConnectedDevices = FindViewById<Spinner>(Resource.Id.spinnerConnectedDevices);
            _adapterConnectedDevices = new ArrayAdapter<string>(this, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerConnectedDevices.Adapter = _adapterConnectedDevices;
            _spinnerChannels = FindViewById<Spinner>(Resource.Id.spinnerChannel);
            ArrayAdapter<Java.Lang.ICharSequence> channelAdapter = ArrayAdapter<Java.Lang.ICharSequence>.CreateFromResource(this, Resource.Array.channels, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerChannels.Adapter = channelAdapter;

            //Get references to button controls
            _btnRefresh = FindViewById<Button>(Resource.Id.btnRefresh);
            _btnRefresh.Click += _btnRefresh_Click;
            _btnConnect = FindViewById<Button>(Resource.Id.btnConnect);
            _btnConnect.Click += _btnConnect_Click;
            _btnMeasure = FindViewById<Button>(Resource.Id.btnMeasure);
            _btnMeasure.Click += _btnMeasure_Click;
            _btnMeasureSequentially = FindViewById<Button>(Resource.Id.btnMeasureSequentially);
            _btnMeasureSequentially.Click += _btnMeasureSequentially_Click;
            _btnMeasureAlternatingly = FindViewById<Button>(Resource.Id.btnMeasureAlternatingly);
            _btnMeasureAlternatingly.Click += _btnMeasureAlternatingly_Click;

            //Get references to textview controls
            _txtPotential = FindViewById<TextView>(Resource.Id.txtPotential);
            _txtCurrent = FindViewById<TextView>(Resource.Id.txtCurrent);
            _txtStatus = FindViewById<TextView>(Resource.Id.txtStatus);
            _txtResults = FindViewById<TextView>(Resource.Id.txtResults);

            InitCAMethod(); //Create the linear sweep voltammetry method that defines the measurement parameters
            InitResultsListView();
        }

        /// <summary>
        /// Called after <c><see cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" /></c> &amp;mdash; or after <c><see cref="M:Android.App.Activity.OnRestart" /></c> when
        /// the activity had been stopped, but is now again being displayed to the
        /// user.
        /// </summary>
        /// <remarks>
        /// <para tool="javadoc-to-mdoc">Called after <c><see cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" /></c> &amp;mdash; or after <c><see cref="M:Android.App.Activity.OnRestart" /></c> when
        /// the activity had been stopped, but is now again being displayed to the
        /// user.  It will be followed by <c><see cref="M:Android.App.Activity.OnResume" /></c>.
        /// </para>
        /// <para tool="javadoc-to-mdoc">
        ///   <i>Derived classes must call through to the super class's
        /// implementation of this method.  If they do not, an exception will be
        /// thrown.</i>
        /// </para>
        /// <para tool="javadoc-to-mdoc">
        ///   <format type="text/html">
        ///     <a href="http://developer.android.com/reference/android/app/Activity.html#onStart()" target="_blank">[Android Documentation]</a>
        ///   </format>
        /// </para>
        /// </remarks>
        /// <since version="Added in API level 1" />
        /// <altmember cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" />
        /// <altmember cref="M:Android.App.Activity.OnStop" />
        /// <altmember cref="M:Android.App.Activity.OnResume" />
        protected override void OnStart()
        {
            base.OnStart();
            //Set the current app context in the psCommSimpleAndroid control
            _psCommSimpleAndroid.CurrentContext = this;
        }

        /// <summary>
        /// Called when app process is resumed and after OnStart when app is started
        /// </summary>
        protected override void OnResume()
        {
            base.OnResume();

            RequestDangerousPermissions(); //Necessary as of Android version 23
        }

        /// <summary>
        /// Request permissions that require user confirmation since android 23
        /// </summary>
        private void RequestDangerousPermissions()
        {
            //Only request permissions on Android versions after 22
            if (Build.VERSION.SdkInt >= BuildVersionCodes.M)
            {
                //Check if location permission is granted, required for BlueTooth
                if (ActivityCompat.CheckSelfPermission(this, Manifest.Permission.AccessFineLocation) != (int)Permission.Granted)
                {
                    //Request location permission
                    ActivityCompat.RequestPermissions(this, new string[] { Manifest.Permission.AccessFineLocation }, (int)PermissionGroupIDs.BlueTooth);
                }
            }
        }

        /// <summary>
        /// Callback received when user permission request has been completed
        /// </summary>
        /// <param name="requestCode"></param>
        /// <param name="permissions"></param>
        /// <param name="grantResults"></param>
        public override void OnRequestPermissionsResult(int requestCode, string[] permissions, [GeneratedEnum] Permission[] grantResults)
        {
            switch ((PermissionGroupIDs)requestCode)
            {
                case PermissionGroupIDs.BlueTooth:
                    if (grantResults.Length == 1 && grantResults[0] == Permission.Granted)
                    {
                        //Permission granted
                        if (!_psCommSimpleAndroid.Connected)
                            DiscoverConnectedDevices(); //Search for BlueTooth devices
                    }
                    else
                    {
                        //Permission denied
                        if (ActivityCompat.ShouldShowRequestPermissionRationale(this, Manifest.Permission.AccessFineLocation))
                        {
                            //Inform user that BlueTooth 
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.", Snackbar.LengthLong).Show();
                        }
                        else
                        {
                            //Occurs when user previously denied permission and selected never ask again
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.\nTo use BlueTooth you need to manualy grant the fine location for this app in the android settings", Snackbar.LengthLong).Show();
                        }
                    }
                    break;
                case PermissionGroupIDs.ExternalStorage: //Not used in this example
                default:
                    base.OnRequestPermissionsResult(requestCode, permissions, grantResults);
                    break;
            }
        }

        /// <summary>
        /// Initializes the data grid view control.
        /// </summary>
        private void InitResultsListView()
        {
            _txtResults.Text = "ID     Time (s)   Current (µA)      Channel";
        }

        /// <summary>
        /// Initializes the CA method.
        /// </summary>
        private void InitCAMethod()
        {
            _methodCA = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCA.Potential = 1f; //Sets the potential at 1V
            _methodCA.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCA.RunTime = 3f; //Sets the measurement runtime at 3s

            _methodCA.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCA.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCA.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCA.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA

            //Sequential multiplexer measurement settings
            _methodCASequential = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCASequential.Potential = 1f; //Sets the potential at 1V
            _methodCASequential.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCASequential.RunTime = 1f; //Sets the measurement runtime at 1s
            _methodCASequential.MuxMethod = MuxMethod.Sequentially; //Set the type of multiplexer measurement
            //Create an array of booleans indicating on which channels you would like to measure
            //For a 16 channel mux you can create a bool array for 16 channels
            bool[] measureOnChannels1 = new bool[8];  //Define bool array for 8 channel multiplexer
            //Specify on which channels to measure            
            measureOnChannels1[0] = false;
            measureOnChannels1[1] = true;
            measureOnChannels1[2] = true;
            measureOnChannels1[3] = false;
            measureOnChannels1[4] = true;
            measureOnChannels1[5] = false;
            measureOnChannels1[6] = false;
            measureOnChannels1[7] = false;
            _methodCASequential.UseMuxChannel = new System.Collections.BitArray(measureOnChannels1); //Set to measure sequentially on channel 1, 3 and 5

            //Alternating multiplexer measurement settings
            //Only Chronoamperometry (Amperometric Detection) and (Open Circuit) Potentiometry support alternating measurements. 
            _methodCAAlternating = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCAAlternating.Potential = 1f; //Sets the potential at 1V
            _methodCAAlternating.IntervalTime = 0.25f; //Sets the sampling interval time at 0.25s (0.25s is the minimum sample time for the multiplexer in alternating mode)
            _methodCAAlternating.RunTime = 2.5f; //Sets the measurement runtime at 2.5s
            _methodCAAlternating.MuxMethod = MuxMethod.Alternatingly; //Set the type of multiplexer measurement
            //Create an array of booleans indicating on which channels you would like to measure
            //For a 16 channel mux you can create a bool array for 16 channels
            bool[] measureOnChannels2 = new bool[8];  //Define bool array for 8 channel multiplexer
            //Specify on which channels to measure
            //For alternating mux measurements only successive channels can be selected starting at the first channel (i.e. 1, 2 and 3 not for example 2, 4 and 7)                  
            measureOnChannels2[0] = true;
            measureOnChannels2[1] = true;
            measureOnChannels2[2] = true;
            measureOnChannels2[3] = false;
            measureOnChannels2[4] = false;
            measureOnChannels2[5] = false;
            measureOnChannels2[6] = false;
            measureOnChannels2[7] = false;
            _methodCAAlternating.UseMuxChannel = new System.Collections.BitArray(measureOnChannels2); //Set to measure alternatingly on channel 1, 2 and 3
        }

        /// <summary>
        /// Displays a Toast message.
        /// </summary>
        /// <param name="message">The message.</param>
        private void DisplayMessage(string message)
        {
            Toast toast = Toast.MakeText(this, message, ToastLength.Short);
            toast.Show();
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the spinner control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            _btnRefresh.Click -= _btnRefresh_Click;
            _btnRefresh.Text = "Refreshing...";
            _btnRefresh.Enabled = false;
            _adapterConnectedDevices.Clear();
            DisplayMessage($"Searching for available devices.");

            try
            {
                _connectedDevices = await _psCommSimpleAndroid.GetConnectedDevices(10000); //Discover connected devices
            }
            catch (Exception ex)
            {
                DisplayMessage(ex.Message);
            }

            foreach (Device d in _connectedDevices)
                _adapterConnectedDevices.Add(d.ToString()); //Add connected devices to spinner control

            int nDevices = _connectedDevices.Length;

            DisplayMessage($"Found {nDevices} device(s).");

            _btnConnect.Enabled = nDevices > 0;
            _btnRefresh.Text = "Refresh";
            _btnRefresh.Enabled = true;
            _btnRefresh.Click += _btnRefresh_Click;
        }

        /// <summary>
        /// Handles the Click event of the _btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private async void _btnRefresh_Click(object sender, System.EventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private async void _btnConnect_Click(object sender, System.EventArgs e)
        {
            _btnConnect.Enabled = false;
            if (!_psCommSimpleAndroid.Connected) //Determine whether a device is currently connected
            {
                if (_adapterConnectedDevices.Count == 0)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    await _psCommSimpleAndroid.Connect(_connectedDevices[_spinnerConnectedDevices.SelectedItemPosition]);
                    DisplayMessage($"Connected to {_psCommSimpleAndroid.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
                finally
                {             
                    //Update UI based on connection status
                    _spinnerConnectedDevices.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnRefresh.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnConnect.Text = _psCommSimpleAndroid.Connected ? "Disconnect" : "Connect";
                    _btnMeasure.Enabled = _psCommSimpleAndroid.Connected;
                    _btnMeasureSequentially.Enabled = _psCommSimpleAndroid.Connected;
                    _btnMeasureAlternatingly.Enabled = _psCommSimpleAndroid.Connected;
                }
            }
            else
            {
                await _psCommSimpleAndroid.Disconnect(); //Disconnect from the connected device
            }
            _btnConnect.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasure control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void _btnMeasure_Click(object sender, EventArgs e)
        {
            _btnMeasure.Enabled = false;
            if (_psCommSimpleAndroid.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    InitResultsListView(); //Reset the data listview
                    //Start measurement defined in the method on the multiplexer channel specified in the channel spinner control
                    _activeMeasurement = await _psCommSimpleAndroid.Measure(_methodCA, _spinnerChannels.SelectedItemPosition);
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            else
            {
                try
                {
                    await _psCommSimpleAndroid.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            _btnMeasure.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the _btnMeasureAlternatingly control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private async void _btnMeasureAlternatingly_Click(object sender, System.EventArgs e)
        {
            _btnMeasureAlternatingly.Enabled = false;
            if (_psCommSimpleAndroid.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    InitResultsListView(); //Reset the data listview
                    //Start the alternating multiplexer measurement on the channels defined in the method
                    _activeMeasurement = await _psCommSimpleAndroid.Measure(_methodCAAlternating);
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            _btnMeasureAlternatingly.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the _btnMeasureSequentially control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private async void _btnMeasureSequentially_Click(object sender, System.EventArgs e)
        {
            _btnMeasureSequentially.Enabled = false;
            if (_psCommSimpleAndroid.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    InitResultsListView(); //Reset the data listview
                    //Start the sequential multiplexer measurement on the channels defined in the method
                    _activeMeasurement = await _psCommSimpleAndroid.Measure(_methodCASequential);
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            _btnMeasureSequentially.Enabled = true;
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Comm.StatusEventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_ReceiveStatus(object sender, PalmSens.Comm.StatusEventArgs e)
        {
            Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
            double potential = status.PotentialReading.Value; //Get the potential
            double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
            PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
            CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

            _txtPotential.Text = $"Potential: {potential.ToString("F3")} V";
            _txtCurrent.Text = $"Current: {currentInRange.ToString("F3")} * {cr.ToString()}";
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CurrentState">State of the current.</param>
        private void _psCommSimpleAndroid_StateChanged(object sender, CommManager.DeviceState CurrentState)
        {
            _txtStatus.Text = $"Status: {CurrentState.ToString()}"; //Updates the device state indicator textbox
            _btnConnect.Enabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            _btnMeasure.Text = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle ? "Measure" : "Abort";
            _btnMeasureSequentially.Enabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            _btnMeasureAlternatingly.Enabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_MeasurementEnded(object sender, Exception e)
        {
            DisplayMessage("Measurement ended.");
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_MeasurementStarted(object sender, EventArgs e)
        {
            DisplayMessage("Measurement started.");
        }

        /// <summary>
        /// Raised when a Simple Curve in the active SimpleMeasurement starts receiving data
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="activeSimpleCurve">The active simple curve.</param>
        private void _psCommSimpleAndroid_SimpleCurveStartReceivingData(object sender, SimpleCurve activeSimpleCurve)
        {
            //Subscribe to the curve's events to receive updates when new data is available and when it iss finished receiving data
            activeSimpleCurve.NewDataAdded += ActiveSimpleCurve_NewDataAdded;
            activeSimpleCurve.CurveFinished += ActiveSimpleCurve_CurveFinished;

            DisplayMessage($"Channel {activeSimpleCurve.MuxChannel + 1}: Curve is receiving new data...");
        }

        /// <summary>
        /// Raised when new data points are added to an active SimpleCurve
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Data.ArrayDataAddedEventArgs"/> instance containing the event data.</param>
        private void ActiveSimpleCurve_NewDataAdded(object sender, PalmSens.Data.ArrayDataAddedEventArgs e)
        {
            if (Looper.MyLooper() != Looper.MainLooper) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                using (var h = new Handler(Looper.MainLooper)) h.Post(() => ActiveSimpleCurve_NewDataAdded(sender, e));
                return;
            }
            SimpleCurve activeCurve = sender as SimpleCurve;
            int startIndex = e.StartIndex; //The index of the first new data point added to the curve
            int count = e.Count; //The number of new data points added to the curve

            for (int i = startIndex; i < startIndex + count; i++)
            {
                double xValue = activeCurve.XAxisValue(i); //Get the value on Curve's X-Axis (potential) at the specified index
                double yValue = activeCurve.YAxisValue(i); //Get the value on Curve's Y-Axis (current) at the specified index
                _txtResults.Text += ($"\n{i + 1}     {xValue.ToString("F2")}     {yValue.ToString("E3")}     {activeCurve.MuxChannel + 1}");
            }
        }

        /// <summary>
        /// Raised when a SimpleCurve stops receiving new data points
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void ActiveSimpleCurve_CurveFinished(object sender, EventArgs e)
        {
            if (Looper.MyLooper() != Looper.MainLooper) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                using (var h = new Handler(Looper.MainLooper)) h.Post(() => ActiveSimpleCurve_CurveFinished(sender, e));
                return;
            }
            SimpleCurve activeCurve = sender as SimpleCurve;

            int nDataPointsReceived = activeCurve.NDataPoints;
            DisplayMessage($"Channel {activeCurve.MuxChannel + 1}: {nDataPointsReceived} data point(s) received.");

            //Unsubscribe from the curves events to avoid memory leaks
            activeCurve.NewDataAdded -= ActiveSimpleCurve_NewDataAdded;
            activeCurve.CurveFinished -= ActiveSimpleCurve_CurveFinished;

            DisplayMessage($"Channel {activeCurve.MuxChannel + 1}: Finished");
        }

        /// <summary>
        /// Raised when a device is disconnected.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CommErrorException">The comm error exception, this only has a value when a disconnect occurred due to a communication error.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private void _psCommSimpleAndroid_Disconnected(object sender, Exception CommErrorException)
        {
            if (CommErrorException != null)
                DisplayMessage(CommErrorException.Message);
            DisplayMessage("Disconnected");

            //Update UI based on connection status
            _spinnerConnectedDevices.Enabled = true;
            _btnRefresh.Enabled = true;
            _btnConnect.Text = "Connect";
            _btnMeasure.Enabled = false;
            _btnMeasure.Text = "Measure";
            _btnMeasureSequentially.Enabled = false;
            _btnMeasureAlternatingly.Enabled = false;
        }
    }
}

