﻿using Android.App;
using Android.OS;
using Android.Support.V7.App;
using Android.Runtime;
using Android.Widget;
using PalmSens.Core.Simplified.Android;
using Android.Views;
using SDKPlot.Android;
using PalmSens.Techniques;
using PalmSens.Devices;
using PalmSens.Core.Simplified.Data;
using PalmSens.Fitting;
using PalmSens.Fitting.Models;
using PalmSens.Comm;
using System;
using PalmSens;
using System.Threading.Tasks;
using Android;
using Android.Support.V4.App;
using Android.Content.PM;
using Android.Support.Design.Widget;
using PalmSens.Techniques.Impedance;

namespace PSSDKPlotEISFitExample
{
    [Activity(Label = "@string/app_name", Theme = "@style/AppTheme", MainLauncher = true)]
    public class MainActivity : Activity
    {
        /// <summary>
        /// The psCommSimpleAndroid control that allows you to control your PalmSens/EmStat device
        /// </summary>
        private PSCommSimpleAndroid _psCommSimpleAndroid;

        /// <summary>
        /// The main view
        /// </summary>
        private View _view;

        /// <summary>
        /// List of permission ids used in the OnRequestPermissionResult
        /// </summary>
        private enum PermissionGroupIDs
        {
            BlueTooth = 0,
            ExternalStorage = 1
        }

        /// <summary>
        /// The refresh devices button
        /// </summary>
        private Button _btnRefresh;

        /// <summary>
        /// The (dis)connect button
        /// </summary>
        private Button _btnConnect;

        /// <summary>
        /// The start/abort measurement button
        /// </summary>
        private Button _btnMeasure;

        /// <summary>
        /// The detect peaks button
        /// </summary>
        private Button _btnFit;

        /// <summary>
        /// The potential textview
        /// </summary>
        private TextView _txtPotential;

        /// <summary>
        /// The current textview
        /// </summary>
        private TextView _txtCurrent;

        /// <summary>
        /// The status textview
        /// </summary>
        private TextView _txtStatus;

        /// <summary>
        /// The potential textview
        /// </summary>
        private TextView _txtRs;

        /// <summary>
        /// The current textview
        /// </summary>
        private TextView _txtRct;

        /// <summary>
        /// The status textview
        /// </summary>
        private TextView _txtCdl;

        /// <summary>
        /// The status textview
        /// </summary>
        private TextView _txtChiSq;

        /// <summary>
        /// The connected devices spinner 
        /// </summary>
        private Spinner _spinnerConnectedDevices;

        /// <summary>
        /// The connected devices list adapter 
        /// </summary>
        private ArrayAdapter<string> _adapterConnectedDevices;

        /// <summary>
        /// The SDKPlot android plot object
        /// </summary>
        private Plot _plot;

        /// <summary>
        /// The instance of method class containing the Cyclic Voltammetry parameters
        /// </summary>
        private ImpedimetricMethod _methodEIS;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement = null;

        /// <summary>
        /// The fit result SimpleCurve
        /// </summary>
        private SimpleCurve[] _fitResultCurves = null;

        /// <summary>
        /// The fit progress instance, reports progress of active fit and allows cancellation
        /// </summary>
        private FitProgress _fitProgress = null;

        /// <summary>
        /// The circuit model class defining the model that will be used for the equivalent circuit fit
        /// </summary>
        private CircuitModel _circuitModel = null;

        /// <summary>
        /// The minimum peak height in µA
        /// </summary>
        private double _minPeakHeight = 0.0001;

        protected override void OnCreate(Bundle bundle)
        {
            base.OnCreate(bundle);

            // Set our view from the "main" layout resource
            SetContentView(Resource.Layout.activity_main);

            _view = FindViewById(Resource.Id.mainLayout);

            //Get reference to psCommSimpleAndroid control
            _psCommSimpleAndroid = FindViewById<PSCommSimpleAndroid>(Resource.Id.pSCommSimpleAndroid);
            _psCommSimpleAndroid.ReceiveStatus += _psCommSimpleAndroid_ReceiveStatus;
            _psCommSimpleAndroid.StateChanged += _psCommSimpleAndroid_StateChanged;
            _psCommSimpleAndroid.MeasurementStarted += _psCommSimpleAndroid_MeasurementStarted;
            _psCommSimpleAndroid.MeasurementEnded += _psCommSimpleAndroid_MeasurementEnded;
            _psCommSimpleAndroid.Disconnected += _psCommSimpleAndroid_Disconnected;

            //Get reference to spinner control
            _spinnerConnectedDevices = FindViewById<Spinner>(Resource.Id.spinnerConnectedDevices);
            _adapterConnectedDevices = new ArrayAdapter<string>(this, Android.Resource.Layout.SimpleSpinnerDropDownItem);
            _spinnerConnectedDevices.Adapter = _adapterConnectedDevices;

            //Get references to button controls
            _btnRefresh = FindViewById<Button>(Resource.Id.btnRefresh);
            _btnRefresh.Click += _btnRefresh_Click;
            _btnConnect = FindViewById<Button>(Resource.Id.btnConnect);
            _btnConnect.Click += _btnConnect_Click;
            _btnMeasure = FindViewById<Button>(Resource.Id.btnMeasure);
            _btnMeasure.Click += _btnMeasure_Click;
            _btnFit = FindViewById<Button>(Resource.Id.btnFit);
            _btnFit.Click += _btnFit_Click;

            //Get references to textview controls
            _txtPotential = FindViewById<TextView>(Resource.Id.txtPotential);
            _txtCurrent = FindViewById<TextView>(Resource.Id.txtCurrent);
            _txtStatus = FindViewById<TextView>(Resource.Id.txtStatus);
            _txtRs = FindViewById<TextView>(Resource.Id.txtRs);
            _txtRct = FindViewById<TextView>(Resource.Id.txtRct);
            _txtCdl = FindViewById<TextView>(Resource.Id.txtCdl);
            _txtChiSq = FindViewById<TextView>(Resource.Id.txtChisq);

            //Get reference to SDKPlot plot control
            _plot = FindViewById<Plot>(Resource.Id.plot);

            InitEISMethod(); //Create the linear sweep voltammetry method that defines the measurement parameters
            InitPlot(); //Resets and initiates the plot control
        }

        /// <summary>
        /// Called after <c><see cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" /></c> &amp;mdash; or after <c><see cref="M:Android.App.Activity.OnRestart" /></c> when
        /// the activity had been stopped, but is now again being displayed to the
        /// user.
        /// </summary>
        /// <remarks>
        /// <para tool="javadoc-to-mdoc">Called after <c><see cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" /></c> &amp;mdash; or after <c><see cref="M:Android.App.Activity.OnRestart" /></c> when
        /// the activity had been stopped, but is now again being displayed to the
        /// user.  It will be followed by <c><see cref="M:Android.App.Activity.OnResume" /></c>.
        /// </para>
        /// <para tool="javadoc-to-mdoc">
        ///   <i>Derived classes must call through to the super class's
        /// implementation of this method.  If they do not, an exception will be
        /// thrown.</i>
        /// </para>
        /// <para tool="javadoc-to-mdoc">
        ///   <format type="text/html">
        ///     <a href="http://developer.android.com/reference/android/app/Activity.html#onStart()" target="_blank">[Android Documentation]</a>
        ///   </format>
        /// </para>
        /// </remarks>
        /// <since version="Added in API level 1" />
        /// <altmember cref="M:Android.App.Activity.OnCreate(Android.OS.Bundle)" />
        /// <altmember cref="M:Android.App.Activity.OnStop" />
        /// <altmember cref="M:Android.App.Activity.OnResume" />
        protected override void OnStart()
        {
            base.OnStart();
            //Set the current app context in the psCommSimpleAndroid control
            _psCommSimpleAndroid.CurrentContext = this;
        }

        /// <summary>
        /// Called when app process is resumed and after OnStart when app is started
        /// </summary>
        protected override void OnResume()
        {
            base.OnResume();

            RequestDangerousPermissions(); //Necessary as of Android version 23
        }

        /// <summary>
        /// Request permissions that require user confirmation since android 23
        /// </summary>
        private void RequestDangerousPermissions()
        {
            //Only request permissions on Android versions after 22
            if (Build.VERSION.SdkInt >= BuildVersionCodes.M)
            {
                //Check if location permission is granted, required for BlueTooth
                if (ActivityCompat.CheckSelfPermission(this, Manifest.Permission.AccessFineLocation) != (int)Permission.Granted)
                {
                    //Request location permission
                    ActivityCompat.RequestPermissions(this, new string[] { Manifest.Permission.AccessFineLocation }, (int)PermissionGroupIDs.BlueTooth);
                }
            }
        }

        /// <summary>
        /// Callback received when user permission request has been completed
        /// </summary>
        /// <param name="requestCode"></param>
        /// <param name="permissions"></param>
        /// <param name="grantResults"></param>
        public override void OnRequestPermissionsResult(int requestCode, string[] permissions, [GeneratedEnum] Permission[] grantResults)
        {
            switch ((PermissionGroupIDs)requestCode)
            {
                case PermissionGroupIDs.BlueTooth:
                    if (grantResults.Length == 1 && grantResults[0] == Permission.Granted)
                    {
                        //Permission granted
                        if (!_psCommSimpleAndroid.Connected)
                            DiscoverConnectedDevices(); //Search for BlueTooth devices
                    }
                    else
                    {
                        //Permission denied
                        if (ActivityCompat.ShouldShowRequestPermissionRationale(this, Manifest.Permission.AccessFineLocation))
                        {
                            //Inform user that BlueTooth 
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.", Snackbar.LengthLong).Show();
                        }
                        else
                        {
                            //Occurs when user previously denied permission and selected never ask again
                            Snackbar.Make(_view, "Without granting permission to fine location BlueTooth will not work in this app.\nTo use BlueTooth you need to manualy grant the fine location for this app in the android settings", Snackbar.LengthLong).Show();
                        }
                    }
                    break;
                case PermissionGroupIDs.ExternalStorage: //Not used in this example
                default:
                    base.OnRequestPermissionsResult(requestCode, permissions, grantResults);
                    break;
            }
        }

        /// <summary>
        /// Initializes the plot control.
        /// </summary>
        private void InitPlot()
        {
            _plot.ClearAll(); //Clear all curves and data from plot
            //Set the Axis labels
            _plot.XAxisLabel = "Log Frequency (Hz)";
            _plot.XAxisType = SDKPlot.AxisType.Logarithmic;
            _plot.YAxisLabel = "Log Z (Ω)";
            _plot.YAxisType = SDKPlot.AxisType.Logarithmic;
            _plot.YAxisSecondaryLabel = "-Phase (°)";
            _plot.YAxisSecondaryType = SDKPlot.AxisType.Linear;
            _plot.AddData("", new double[0], new double[0]); //Add a empty data array to draw an empty plot
        }

        /// <summary>
        /// Initializes the EIS method.
        /// </summary>
        private void InitEISMethod()
        {
            _methodEIS = new ImpedimetricMethod(); 
            _methodEIS.EquilibrationTime = 0f; //Equilabrates the cell at the defined potential for 5 seconds before starting the measurement
            _methodEIS.Potential = 0f; //Sets the dc potential
            _methodEIS.Eac = 0.01f; // Sets the ac potential at 10 mV RMS
            _methodEIS.ScanType = enumScanType.Fixed; //Single measurement at fixed frequency
            _methodEIS.FreqType = enumFrequencyType.Scan; //Scan a range of frequencies
            _methodEIS.MaxFrequency = 200000; //Start scan at 200 000 Hz
            _methodEIS.MinFrequency = 100; //End scan at 100 Hz
            _methodEIS.nFrequencies = 50; //Sets number of frequencies in scan to 50

            _methodEIS.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr10mA); //Starts equilabration in the 10mA current range
            _methodEIS.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Min current range 1µA
            _methodEIS.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr10mA); //Max current range 10mA
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the spinner control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            _btnRefresh.Click -= _btnRefresh_Click;
            _btnRefresh.Text = "Refreshing...";
            _btnRefresh.Enabled = false;
            _adapterConnectedDevices.Clear();
            DisplayMessage($"Searching for available devices.");

            try
            {
                _connectedDevices = await _psCommSimpleAndroid.GetConnectedDevices(10000); //Discover connected devices
            }
            catch (Exception ex)
            {
                DisplayMessage(ex.Message);
            }

            foreach (Device d in _connectedDevices)
                _adapterConnectedDevices.Add(d.ToString()); //Add connected devices to spinner control

            int nDevices = _connectedDevices.Length;

            DisplayMessage($"Found {nDevices} device(s).");

            _btnConnect.Enabled = nDevices > 0;
            _btnRefresh.Text = "Refresh";
            _btnRefresh.Enabled = true;
            _btnRefresh.Click += _btnRefresh_Click;
        }

        /// <summary>
        /// Displays a Toast message.
        /// </summary>
        /// <param name="message">The message.</param>
        private void DisplayMessage(string message)
        {
            Toast toast = Toast.MakeText(this, message, ToastLength.Short);
            toast.Show();
        }

        /// <summary>
        /// Handles the Click event of the _btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private async void _btnRefresh_Click(object sender, System.EventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private async void _btnConnect_Click(object sender, System.EventArgs e)
        {
            _btnConnect.Enabled = false;
            if (!_psCommSimpleAndroid.Connected) //Determine whether a device is currently connected
            {
                if (_adapterConnectedDevices.Count == 0)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    await _psCommSimpleAndroid.Connect(_connectedDevices[_spinnerConnectedDevices.SelectedItemPosition]);
                    DisplayMessage($"Connected to {_psCommSimpleAndroid.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
                finally
                {
                    //Update UI based on connection status
                    _spinnerConnectedDevices.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnRefresh.Enabled = !_psCommSimpleAndroid.Connected;
                    _btnConnect.Text = _psCommSimpleAndroid.Connected ? "Disconnect" : "Connect";
                    _btnMeasure.Enabled = _psCommSimpleAndroid.Connected;
                }
            }
            else
            {
                await _psCommSimpleAndroid.Disconnect(); //Disconnect from the connected device
            }
            _btnConnect.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasure control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void _btnMeasure_Click(object sender, EventArgs e)
        {
            _btnMeasure.Enabled = false;
            if (_psCommSimpleAndroid.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    _plot.ClearAll(); //Clears data from previous measurements from the plot
                    _activeMeasurement = await _psCommSimpleAndroid.Measure(_methodEIS); //Start measurement defined in the method

                    //Add bode curves to plot
                    _plot.AddSimpleCurve(_activeMeasurement.NewSimpleCurve(PalmSens.Data.DataArrayType.Frequency, PalmSens.Data.DataArrayType.Z, "Z Abs")[0]);
                    _plot.AddSimpleCurve(_activeMeasurement.NewSimpleCurve(PalmSens.Data.DataArrayType.Frequency, PalmSens.Data.DataArrayType.Phase, "-Phase")[0], true);
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            else
            {
                try
                {
                    await _psCommSimpleAndroid.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    DisplayMessage(ex.Message);
                }
            }
            _btnMeasure.Enabled = true;
        }

        /// <summary>
        /// Handles the Click event of the _btnPeaks control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private async void _btnFit_Click(object sender, EventArgs e)
        {
            if (_fitProgress == null)
            {
                FitEquivalentCircuit();
                _btnFit.Text = "Cancel fit";
            }
            else _fitProgress.Cancel();
        }
        /// <summary>
         /// Fits the equivalent circuit.
         /// </summary>
         /// <exception cref="System.NotImplementedException"></exception>
        private async Task FitEquivalentCircuit()
        {
            _circuitModel = BuildEquivalentCircuitModel(); //Build the model to fit on the data
            FitOptionsCircuit fitOptions = InitFitOptions(_circuitModel); //Define the options for the fit
            FitAlgorithm fitAlgorithm = FitAlgorithm.FromAlgorithm(fitOptions); //Set up the instance of the fit class
            _fitProgress = new FitProgress(); //Create an instance of the peak detection progress reporter (this can also be used to cancel the peak detection)
            _fitProgress.ProgressChanged += FitProgress_ProgressChanged;
            await fitAlgorithm.ApplyFitCircuitAsync(_fitProgress); //Start the fit
        }

        /// <summary>
        /// Builds the equivalent circuit model.
        /// </summary>
        /// <returns></returns>
        private CircuitModel BuildEquivalentCircuitModel()
        {
            CircuitModel circuitModel = new CircuitModel();
            circuitModel.SetEISdata(_activeMeasurement.Measurement.EISdata[0]); //Sets reference to measured data
            circuitModel.SetCircuit("R(RC)"); //Sets the circuit defined in the CDC code string, in this case a Randles circuit
            circuitModel.SetInitialParameters(
                new double[] {
                    100, //The initial value for the solution resistance (series resistor)
                    8000, //The initial value for the charge transfer resistance (parallel resistor)
                    1e-8 //The initial value for the double layer capacitance (parallel capacitor)
                });
            //Setting the initial parameters is recommended to ensure a good quality fit and 
            //reduce to the time it takes to find the fit

            return circuitModel;
        }

        /// <summary>
        /// Initializes the fit options.
        /// </summary>
        /// <param name="circuitModel">The circuit model.</param>
        /// <returns></returns>
        private FitOptionsCircuit InitFitOptions(CircuitModel circuitModel)
        {
            FitOptionsCircuit fitOptions = new FitOptionsCircuit();
            fitOptions.Model = circuitModel;
            fitOptions.RawData = _activeMeasurement.Measurement.EISdata[0]; //The measured data to fit onto
            fitOptions.SelectedAlgorithm = Algorithm.LevenbergMarquardt;
            fitOptions.MaxIterations = 100; //The maximum number of iterations, 500 by default
            fitOptions.MinimumDeltaErrorTerm = 1e-9; //The minimum delta in the error term (sum of squares difference between model and data), default is 1e-9      
            fitOptions.EnableMinimunDeltaErrorTerm = true; //Enable the minimum delta error as a stop condition within the fit algorithm, default is true
            fitOptions.MinimumDeltaParameters = 1e-12; //The minimum delta parameter step size, default is 1e-12
            fitOptions.EnableMinimunDeltaParameters = true; //Enable the minimum delta parameter step size as a stop condition within the fit algorithm, default is true
            fitOptions.Lambda = 1e-2; //The starting value for the Levenberg Marquardt scaling factor, default is 1e-2
            fitOptions.LambdaFactor = 10; //The scaling value for the Levenberg Marquardt scaling factor, default is 10

            return fitOptions;
        }

        /// <summary>
        /// Raised when fit progress changed.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The e.</param>
        /// <exception cref="NotImplementedException"></exception>
        private void FitProgress_ProgressChanged(object sender, FitProgressUpdate e)
        {
            switch (e.Progress)
            {
                case EnumFitProgress.Started:
                    DisplayMessage("Started circuit fit.");
                    _txtRs.Text = "Rs: - Ω, %error: - %";
                    _txtRct.Text = "Rct: - Ω, %error: - %";
                    _txtCdl.Text = "Cdl: - F, %error: - %";
                    _txtChiSq.Text = "Chi-sq: - , Iterations: -";
                    break;
                case EnumFitProgress.FitIterated:
                    _txtChiSq.Text = $"Chi-sq: - , Iterations: {e.NIterations.ToString()}"; 
                    break;
                case EnumFitProgress.Cancelled:
                case EnumFitProgress.Finished:
                    SetFitResults(e.Result);
                    DisplayMessage("Circuit fit completed.");
                    _fitProgress.ProgressChanged -= FitProgress_ProgressChanged;
                    _fitProgress = null;
                    _btnFit.Text = "Fit circuit";
                    break;
            }
        }

        /// <summary>
        /// Sets the fit results.
        /// </summary>
        /// <param name="result">The result.</param>
        private void SetFitResults(FitResult result)
        {
            //Add results to console
            _txtRs.Text = $"Rs: {result.FinalParameters[0].ToString("0.00E+000")} Ω, %error: {result.ParameterSDs[0].ToString("0.00E+000")} %";
            _txtRct.Text = $"Rct: {result.FinalParameters[1].ToString("0.00E+000")} Ω, %error: {result.ParameterSDs[1].ToString("0.00E+000")} %";
            _txtCdl.Text = $"Cdl: {result.FinalParameters[2].ToString("0.00E+000")} F, %error: {result.ParameterSDs[2].ToString("0.00E+000")} %";
            _txtChiSq.Text = $"Chi-sq: {result.ChiSq.ToString("0.00E+000")} , Iterations: {result.NIterations - 1}"; //Excludes initial guess as iteration
            switch (result.ExitCode)
            {
                case ExitCodes.MinimumDeltaErrorTerm:
                    _txtChiSq.Text += ", Reached min delta error sum squares";
                    break;
                case ExitCodes.MinimumDeltaParameters:
                    _txtChiSq.Text += ", Reached min delta parameter step size";
                    break;
                case ExitCodes.MaxIterations:
                    _txtChiSq.Text += ", Reached max iterations";
                    break;
                case ExitCodes.HessianNonPositive:
                    _txtChiSq.Text += ", Error determing step for next iteration";
                    break;
            }

            //Remove previous fit results from plot
            if (_fitResultCurves != null)
                foreach (SimpleCurve sc in _fitResultCurves)
                    if (_plot.ContainsSimpleCurve(sc))
                        _plot.RemoveSimpleCurve(sc);

            //Generate new simple curves based on fit results
            _circuitModel.SetInitialParameters(result.FinalParameters); //Update the initial parameters with fit results
            _fitResultCurves = new SimpleCurve[2];
            _fitResultCurves[0] = new SimpleCurve(_circuitModel.GetCurveZabsOverFrequency(false)[0], _activeMeasurement);
            _fitResultCurves[0].Title = "Z Abs Fit";
            _fitResultCurves[1] = new SimpleCurve(_circuitModel.GetCurvePhaseOverFrequency(false)[0], _activeMeasurement);
            _fitResultCurves[1].Title = "-Phase Fit";

            //Add results to plot   
            _plot.AddSimpleCurve(_fitResultCurves[0], false, false);
            _plot.AddSimpleCurve(_fitResultCurves[1], true, true);
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Comm.StatusEventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_ReceiveStatus(object sender, PalmSens.Comm.StatusEventArgs e)
        {
            Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
            double potential = status.PotentialReading.Value; //Get the potential
            double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
            PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
            CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

            _txtPotential.Text = $"Potential: {potential.ToString("F3")} V";
            _txtCurrent.Text = $"Current: {currentInRange.ToString("F3")} * {cr.ToString()}";
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CurrentState">State of the current.</param>
        private void _psCommSimpleAndroid_StateChanged(object sender, CommManager.DeviceState CurrentState)
        {
            _txtStatus.Text = $"Status: {CurrentState.ToString()}"; //Updates the device state indicator textbox
            _btnConnect.Enabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            _btnMeasure.Text = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle ? "Measure" : "Abort";
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_MeasurementStarted(object sender, EventArgs e)
        {
            DisplayMessage("Measurement started.");
            _btnFit.Enabled = false;
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _psCommSimpleAndroid_MeasurementEnded(object sender, Exception e)
        {
            DisplayMessage("Measurement ended.");
            _btnFit.Enabled = _activeMeasurement.Measurement.EISdata.Count > 0 && _activeMeasurement.Measurement.EISdata[0].NPoints > 0;
        }

        /// <summary>
        /// Raised when a device is disconnected.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CommErrorException">The comm error exception, this only has a value when a disconnect occurred due to a communication error.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private void _psCommSimpleAndroid_Disconnected(object sender, Exception CommErrorException)
        {
            if (CommErrorException != null)
                DisplayMessage(CommErrorException.Message); 
            DisplayMessage("Disconnected");

            //Update UI based on connection status
            _spinnerConnectedDevices.Enabled = true;
            _btnRefresh.Enabled = true;
            _btnConnect.Text = "Connect";
            _btnMeasure.Enabled = false;
            _btnMeasure.Text = "Measure";
        }
    }
}